#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned int uint;

struct dirinfo {
  unsigned char status;
  unsigned short sec_len;
  unsigned short start_sec;
  char file_name[8];
  char file_ext[4];
  unsigned short dir_no;
};

typedef struct dirinfo dirinfo;

int  getdirinfo (dirinfo *dir, int);
int  readdirtable();

long SECTOR_LEN = 128;
long VTOC_SEC   = 360 - 1;
long DIR_START  = 361 - 1;
long atr_offset = 0;

unsigned char vtoc[128], dirtable[1024];
unsigned char bittab[8] = {1, 2, 4, 8, 16, 32, 64, 128};
unsigned char buffer0[256], buffer1[256];

FILE *fhand;
short freesec, allsec;
char file_name[100];


int main(int argc, char **argv)
{
  int i, j, arg;
  dirinfo dir;
  unsigned char c;
  char *str;

  arg = 1;
  if(*argv[1] == '-') arg = 2;

  str = argv[arg];
  j = strlen(str);
  for(i = 0; i < j; i++)
    if(str[i] == '.') break;
  i++;
  if(strcmp(str + i, "xfd") == 0)
    atr_offset = 0;
  else if(strcmp(str + i, "atr") == 0)
    atr_offset = 16;
  else
  {
    error();
    return 20;
  }

  if ((fhand = fopen(argv[arg], "r+b")) == NULL)
  {
    fprintf(stderr, "Can't open file: %s \n", argv[arg]);
    return 5;
  }

  if(arg != 1)
  {
    switch(argv[1][1]) {
      case 'i': import_file(argv[arg + 1]); break;
      case 'e': export_file(argv[arg + 1]); break;
      case 'd': delete_file(argv[arg + 1]); break;
      case 'p': printdir(); break;
    }
  }
  else while (c != 'X')
  {
     printf("*** XFD-DOS V1.0 by R. Scholz\n");
     printf("1) Directory \n");
     printf("2) Xfd ==> DOS \n");
     printf("3) DOS ==> Xfd \n");
     printf("4) Delete from Xfd \n");
     printf("X) Exit \n\n");
     printf("Please select: \n\n");

     c = toupper(getch());
     switch(c) {
	case '1': printdir(); break;
	case '2': export(); break;
	case '3': import(); break;
	case '4': delete(); break;
	case 'T': do_test(); break;
	default : break;
     }
  }

  fclose(fhand);
  return 0;
}

int error()
{
  printf("Usage: xfddos [-deip] <file.XFD | ATR> \n");
  printf("               -d delete file \n");
  printf("               -e export file \n");
  printf("               -i import file \n");
  printf("               -p print directory \n");

  return 1;
}

int do_test()
{
  int i;
  unsigned char buf[256];

  read_sector(4, buf);
  read_sector(5, buf);
  read_sector(6, buf);

  return 0;
}

int export()
{
  printf("Enter file to export:");
  scanf("%s", file_name);
  export_file(file_name);
  return 1;
}

int import()
{
  printf("Enter file to import:");
  scanf("%s", file_name);
  import_file(file_name);
  return 1;
}

int delete()
{
  printf("Enter file to delete:");
  scanf("%s", file_name);
  delete_file(file_name);
  return 1;
}

int printdir()
{
  int i, j;
  dirinfo d;

  for(i = 0, j = 0; i < 64; i++)
  {
    getdirinfo(&d, i);
    if (d.status != '*')
    {
       j++;
       printf("%2.2u %c %-8.8s %-3.3s %3.3u\n",
	      i, d.status, d.file_name, d.file_ext, d.sec_len);
       if (!(j % 20)) getch();
    }
  }
  readvtoc();
  allsec = vtoc[1] + (vtoc[2] << 8);
  freesec = vtoc[3] + (vtoc[4] << 8);
  printf("\n%u of %u sectors free\n", freesec, allsec);
  getch();
  return 1;
}

int getdirinfo(dirinfo *dir, int dir_no)
{
  int i, index;

  if (dir_no < 0 || dir_no > 63) return 0;

  index = dir_no * 16;
  readdirtable();

  memcpy(dir, &(dirtable[index]), 16);

  switch (dirtable[index] & 253)
  {
    case   1: dir -> status = 'O'; break;
    case  32: dir -> status = 'L'; break;
    case  64: dir -> status = ' '; break;
    case 128: dir -> status = 'D'; break;
    default : dir -> status = '*'; break;
  }

  dir -> file_ext[3] = 0;
  dir -> dir_no = dir_no;
  return 1;
}

int readdirtable()
{
  fseek(fhand, atr_offset + DIR_START * SECTOR_LEN, 0);
  fread(dirtable, 1, 1024, fhand);
  return 1;
}

int writedirtable()
{
  fseek(fhand, atr_offset + DIR_START * SECTOR_LEN, 0);
  fwrite(dirtable, 1, 1024, fhand);
  return 1;
}

int readvtoc()
{
  int i;

  fseek(fhand, atr_offset + VTOC_SEC * SECTOR_LEN, SEEK_SET);
  fread(vtoc, 1, SECTOR_LEN, fhand);
  return 1;
}

int writevtoc()
{
  int i;

  fseek(fhand, atr_offset + VTOC_SEC * SECTOR_LEN, SEEK_SET);
  fwrite(vtoc, 1, SECTOR_LEN, fhand);
  return 1;
}

int getexistdir(dirinfo *dir, char *name)
{
  int i;

  for(i = 0; i < 64; i++)
  {
    getdirinfo(dir, i);
    if (strcmp(dir->file_name, name) == 0 && dir->status == ' ') break;
  }
  if (i == 64) return -1;
  return i;
}

int getfreedir(dirinfo *dir)
{
  int i;

  for(i = 0; i < 64; i++)
  {
    getdirinfo(dir, i);
    if (dir -> status == 'D' || dir -> status == '*') break;
  }
  if (i == 64) return -1;
  return i;
}

int clearbit(unsigned char *c, int bit_no)
{
  unsigned char op;

  op = ~bittab[bit_no];
  *c = *c & op;
  return 1;
}

int setbit(unsigned char *c, int bit_no)
{
  unsigned char op;

  op = bittab[bit_no];
  *c = *c | op;
  return 1;
}

int testbit(unsigned char c, int bit_no)
{
  unsigned char op;

  op = bittab[bit_no];
  if ((c & op) == 0) return 0;
  else return 1;
}

/* modus = 'O' == Sector occupied */
/* Sector Free sonst              */
int mark_sector(uint sector, char modus)
{
  long byte;
  int bit_no;

  byte = sector / 8;
  bit_no = 7 - (sector % 8);
  if (modus == 'O')
    clearbit(&(vtoc[byte + 10]), bit_no);
  else
    setbit(&(vtoc[byte + 10]), bit_no);

  return 1;
}

long next_free_sector()
{
  uint i;

  for (i = 1; i < 360; i++)
    if (free_sector(i)) return i;
  for (i = 369; i < 720; i++)
    if (free_sector(i)) return i;

  return 0;
}

int free_sector(uint sector)
{
  long byte;
  int bit_no;

  byte = sector / 8;
  bit_no = 7 - (sector % 8);
  return (testbit(vtoc[byte+10], bit_no));
}

int export_file(char *export_name)
{
  char name[12];
  uint sector;
  dirinfo dir;
  FILE *fout;

  atari_name(export_name, name);

  if(getexistdir(&dir, name) == -1)
  {
    printf("File %s not in Directory! \n", export_name);
    return 0;
  }

  if ((fout = fopen(export_name, "wb")) == NULL) return 0;

  for(sector = dir.start_sec; sector != 0;)
  {
    read_sector(sector, buffer0);
    fwrite(buffer0, 1, buffer0[SECTOR_LEN-1], fout);
    sector = buffer0[SECTOR_LEN-2] + ((buffer0[SECTOR_LEN-3] & 3) << 8);
  }

  fclose(fout);
  return 1;
}

int read_sector(uint sector, unsigned char *buffer)
{
  fseek(fhand, atr_offset + (long)((sector - 1) * SECTOR_LEN), 0);
  fread(buffer, 1, SECTOR_LEN, fhand);
  return 1;
}

int write_sector(uint sector, unsigned char *buffer)
{
  fseek(fhand, atr_offset + (long)((sector - 1) * SECTOR_LEN), 0);
  fwrite(buffer, 1, SECTOR_LEN, fhand);
  return 1;
}

int atari_name(char *pc_name, char *atari_name)
{
  int i, j, k;

  j = strlen(pc_name);

  for(i = j - 1; i >= 0; i--)
    if(pc_name[i] == '\\') break;

  memset(atari_name, 32, 12);

  for(k = 0, i++; i < j; i++, k++)
    if (pc_name[i] == '.') break;
    else atari_name[k] = (char)toupper(pc_name[i]);

  for(k = 8; i < j; i++, k++)
  {
     atari_name[k] = (char)toupper(pc_name[i+1]);
  }
  atari_name[11] = 0;
  return 1;
}

int import_file(char *import_name)
{
  int dir_no, flen;
  char name[12];
  unsigned char *dindex;
  uint i, sector, sector_old, j, len;
  dirinfo dir;
  FILE *fin;

  atari_name(import_name, name);

  printf("%s %s \n", name, import_name);

  if ((fin = fopen(import_name, "rb")) == NULL)
  {
    printf("Can't open file %s on PC! \n", import_name);
    return 0;
  }

  readdirtable();
  if (getexistdir(&dir, name) != -1)
  {
     delete_file(import_name);
  }

  readdirtable();
  readvtoc();

  dir_no = getfreedir(&dir);
  if (dir_no == -1)
  {
    printf("No space in directory left! \n");
    return 0;
  }

  printf("import DIR:%u \n", dir_no);

  for(i = 0, flen = 0; ;i++)
  {
    len = fread(buffer0, 1, SECTOR_LEN - 3, fin);
    flen += len;
    buffer0[SECTOR_LEN-1] = len;
    buffer0[SECTOR_LEN-2] = 0;
    buffer0[SECTOR_LEN-3] = dir_no << 2;

    sector = next_free_sector();
    /* printf("s:%u ", sector); */
    if(sector == 0) break;

    if(i == 0)
    {
      dir.start_sec = sector;
      dir.status = 66;
      mark_sector(sector, 'O');
      memcpy(buffer1, buffer0, SECTOR_LEN);
    }
    else
    {
      buffer1[SECTOR_LEN-3] = (dir_no << 2) | ((sector >> 8) & 3);
      buffer1[SECTOR_LEN-2] = sector & 255;
      mark_sector(sector, 'O');
      write_sector(sector_old, buffer1);
      memcpy(buffer1, buffer0, SECTOR_LEN);
    }
    sector_old = sector;
    if (len != SECTOR_LEN-3) break;
  }

  fclose(fin);

  dir.sec_len = i + 1;
  dindex = dirtable + dir_no * 16;
  dindex[0] = 66;
  dindex[1] = dir.sec_len & 255;
  dindex[2] = dir.sec_len >> 8;
  dindex[3] = dir.start_sec & 255;
  dindex[4] = dir.start_sec >> 8;
  strncpy(dindex + 5, name, 11);

  /* Anzahl freie Sektoren lesen, benutzte abziehen und speichern        */
  freesec = vtoc[3] + (vtoc[4] << 8);
  freesec -= dir.sec_len;
  vtoc[3] = freesec & 255;
  vtoc[4] = freesec >> 8;

  if (sector != 0)
  {
    mark_sector(sector, 'O');
    write_sector(sector, buffer1);
    /* printf("Import File Length: %u %u\n", dir.sec_len, flen);*/
  }

  if((i == 0 && sector != NULL) || (i > 0))
  {
    writevtoc();
    writedirtable();
  }

  return 1;
}

int delete_file(char *pc_name)
{
  int dir_no;
  char name[12];
  uint i, sector;
  dirinfo dir;

  readvtoc();
  atari_name(pc_name, name);

  dir_no = getexistdir(&dir, name);
  if (dir_no == -1)
  {
     printf ("file %s not in directory!\n", pc_name);
     return 0;
  }

  /* Directory als geloescht kennzeichenen                                */
  dirtable[dir_no * 16] = 128;

  for (sector = dir.start_sec, i = 0; sector != 0; i++)
  {
     read_sector(sector, buffer0);
     mark_sector(sector, 'F');
     sector = buffer0[SECTOR_LEN-2] + ((buffer0[SECTOR_LEN-3] & 3) << 8);
  }
  /* printf("\n"); */

  /* Anzahl freie Sektoren lesen, geloeschte addieren und speichern        */
  freesec = vtoc[3] + (vtoc[4] << 8);
  freesec += i;
  vtoc[3] = freesec & 255;
  vtoc[4] = freesec >> 8;

  writevtoc();
  writedirtable();
  return 1;
}
